/*
 * Implementacao do algoritmo: Recursive MLSL-AP-EF
 *
 * As entradas sao:
 * - u(n), reference signals
 *  - d(n), desired signals
 *
 * As saidas sao:
 * - e(n)=d(n)-d^(n|Un), the a posteriori estimation error
 *
 * Paulo Lopes 24-2-2007
 */
#include "global.h"
#include "MLSLv3.h"
#include <stdlib.h>

MLSLattice::MLSLattice(int inSizeAr, int outSizeAr, int orderAr, double lambdaAr, double delta)
  :McLatticePr(inSizeAr, outSizeAr, orderAr)
{
	int m,i,j;

	lambda=(MyData)lambdaAr;

	Bi1p=(MyData *)calloc(inSizeSq*(order+1), sizeof(MyData));
	Fi1p=(MyData *)calloc(inSizeSq*(order+1), sizeof(MyData));
	GammaF1p=(MyData *)calloc(inSizeSq*(order+1), sizeof(MyData));
	GammaB1p=(MyData *)calloc(inSizeSq*(order+1), sizeof(MyData));
	gamma1p=(MyData *)calloc(order+1, sizeof(MyData));
	kappa1p=(MyData *)calloc(outSize*inSize*(order+1), sizeof(MyData));

	Bi0p=(MyData *)calloc(inSizeSq*(order+1), sizeof(MyData));
	Fi0p=(MyData *)calloc(inSizeSq*(order+1), sizeof(MyData));

	for(m=0; m<=order; m++) {
		for(i=0; i<inSize; i++)
			eta0(m,i)=0;
		for(i=0; i<inSize; i++)
			psi0(m,i)=0;

		for(i=0; i<inSize; i++){
			for(j=0; j<inSize; j++)
				Bi0(m,i,j)=0;}
		for(i=0; i<inSize; i++)
			Bi0(m,i,i)=(MyData)(1/delta);

		for(i=0; i<inSize; i++){
			for(j=0; j<inSize; j++)
				Fi0(m,i,j)=0;}
		for(i=0; i<inSize; i++)
			Fi0(m,i,i)=(MyData)(1/delta);

		for(i=0; i<inSize; i++){
			for(j=0; j<inSize; j++)
				GammaF0(m,i,j)=0;}
		for(i=0; i<inSize; i++){
			for(j=0; j<inSize; j++)
				GammaB0(m,i,j)=0;}
		gamma0(m)=1;
		for(i=0; i<outSize; i++)
			alfa0(m,i)=0;
		for(i=0; i<outSize; i++) {
			for(j=0; j<inSize; j++)
				kappa0(m,i,j)=0;
		}
	}
}

MLSLattice::~MLSLattice(void)
{
	free(Bi1p);
	free(Fi1p);
	free(GammaF1p);
	free(GammaB1p);
	free(gamma1p);
	free(kappa1p);

	free(Bi0p);
	free(Fi0p);
}

void MLSLattice::Update(val_type *in, val_type *d)
{
	int m,i,j,k,l;
	MyData aux1, aux2, aux3;

	//	Inicializa�o
	
	for(i=0; i<inSize; i++)
		eta1(0,i)=(MyData)in[i];

	for(i=0; i<inSize; i++)
		psi1(0,i)=(MyData)in[i];

	for(i=0; i<outSize; i++)
		alfa1(0,i)=(MyData)d[i];

	gamma1(0)=1;

	for(m=1; m<=order; m++){
		
		// eta1[m] = eta1[m - 1] + GammaF0[m].psi0[m - 1];

		for(i=0; i<inSize; i++) {
			eta1(m,i)=eta1(m-1,i);
			for(k=0; k<inSize; k++)
				eta1(m,i)+=GammaF0(m,i,k)*psi0(m-1,k);
		}
	
		// psi1[m] = psi0[m - 1] + GammaB0[m].eta1[m - 1];

		for(i=0; i<inSize; i++) {
			psi1(m,i)=psi0(m-1,i);
			for(k=0; k<inSize; k++)
				psi1(m,i)+=GammaB0(m,i,k)*eta1(m-1,k);
		}

		// Fi1[m]  = (Fi0[m] - (Fi0[m].eta1[m - 1].Transpose(eta1[m - 1]).Fi0[m]*gamma0[m - 1])/
		//			 (lambda + Transpose(eta1[m - 1]).Fi0[m].eta1[m - 1]*gamma0[m - 1]))/lambda;

		aux1=0;
		for(k=0; k<inSize; k++) {
			aux2=0;
			for(l=0; l<inSize; l++)
				aux2+=Fi0(m,k,l)*eta1(m-1,l);
			aux1+=aux2*eta1(m-1,k);
		}
		assert(aux1>=0);
		
		aux1=1/(lambda+gamma0(m-1)*aux1);

		for(i=0; i<inSize; i++) {
			for(j=0; j<inSize; j++) {
				aux2=0;
				for(k=0; k<inSize; k++)
					aux2+=Fi0(m,i,k)*eta1(m-1,k);
				aux3=0;
				for(l=0; l<inSize; l++)
					aux3+=eta1(m-1,l)*Fi0(m,l,j);
				Fi1(m,i,j)=(Fi0(m,i,j)-aux2*aux3*gamma0(m-1)*aux1)/lambda;
			}
			assert(Fi1(m,i,i)>0);
		}
		
		// Bi1[m]  = (Bi0[m] - (Bi0[m].psi1[m - 1].Transpose(psi1[m - 1]).Bi0[m]*gamma1[m - 1])/
		//			 (lambda + Transpose(psi1[m - 1]).Bi0[m].psi1[m - 1]*gamma1[m - 1]))/lambda;

		aux1=0;
		for(k=0; k<inSize; k++) {
			aux2=0;
			for(l=0; l<inSize; l++)
				aux2+=Bi0(m,k,l)*psi1(m-1,l);
			aux1+=aux2*psi1(m-1,k);
		}
		assert(aux1>=0);
		
		aux1=1/(lambda+gamma1(m-1)*aux1);

		for(i=0; i<inSize; i++) {
			for(j=0; j<inSize; j++) {
				aux2=0;
				for(k=0; k<inSize; k++)
					aux2+=Bi0(m,i,k)*psi1(m-1,k);
				aux3=0;
				for(l=0; l<inSize; l++)
					aux3+=psi1(m-1,l)*Bi0(m,l,j);
				Bi1(m,i,j)=(Bi0(m,i,j)-aux2*aux3*gamma1(m-1)*aux1)/lambda;
			}
			assert(Bi1(m,i,i)>0);
		}

		// GammaF1[m] = GammaF0[m] - eta1[m].Transpose(psi0[m - 1]).Bi0[m]*gamma0[m - 1];

		for(i=0; i<inSize; i++) {
			for(j=0; j<inSize; j++) {
				aux1=0;
				for(k=0; k<inSize; k++)
					aux1+=psi0(m-1,k)*Bi0(m,k,j);
				GammaF1(m,i,j)=GammaF0(m,i,j)-eta1(m,i)*aux1*gamma0(m-1);
			}
		}
		
		// GammaB1[m] = GammaB0[m] - psi1[m].Transpose(eta1[m - 1]).Fi1[m]*gamma0[m - 1];

		for(i=0; i<inSize; i++) {
			for(j=0; j<inSize; j++) {
				aux1=0;
				for(k=0; k<inSize; k++)
					aux1+=eta1(m-1,k)*Fi1(m,k,j);
				GammaB1(m,i,j)=GammaB0(m,i,j)-psi1(m,i)*aux1*gamma0(m-1);
			}
		}
	
		// gamma1[m]  = gamma1[m - 1] - Transpose(psi1[m - 1]).Bi1[m].psi1[m - 1]*(gamma1[m - 1]*gamma1[m - 1]);

		aux1=0;
		for(k=0; k<inSize; k++){
			aux2=0;
			for(l=0; l<inSize; l++) 
				aux2+=Bi1(m,k,l)*psi1(m-1,l);
			aux1+=aux2*psi1(m-1,k);
		}
		assert(aux1>=0);

		gamma1(m)=gamma1(m-1)-aux1*gamma1(m-1)*gamma1(m-1);
		assert(gamma1(m)>0);
		assert(gamma1(m)<=1);
		
		// alfa1[m]  = alfa1[m - 1] - kappa0[m].psi1[m - 1]

		for(i=0; i<outSize; i++) {
			alfa1(m,i)=alfa1(m-1,i);
			for(k=0; k<inSize; k++)
				alfa1(m,i)-=kappa0(m,i,k)*psi1(m-1,k);
		}
		
		// kappa1[m] = kappa0[m] + alfa1[m].Transpose(psi1[m - 1]).Bi1[m]*gamma1[m - 1];

		for(i=0; i<outSize; i++) {
			for(j=0; j<inSize; j++) {
				aux1=0;
				for(k=0; k<inSize; k++)
					aux1+=psi1(m-1,k)*Bi1(m,k,j);
				kappa1(m,i,j)=kappa0(m,i,j)+alfa1(m,i)*aux1*gamma1(m-1);
			}
		}
	}

	// time update

	for(m=0; m<=order; m++){
		
		for(i=0; i<inSize; i++)
			eta0(m,i)=eta1(m,i);
		
		for(i=0; i<inSize; i++)
			psi0(m,i)=psi1(m,i);
		
		for(i=0; i<inSize; i++) {
			for(j=0; j<inSize; j++)
				Fi0(m,i,j)=Fi1(m,i,j);
		}
		
		for(i=0; i<inSize; i++) {
			for(j=0; j<inSize; j++)
				Bi0(m,i,j)=Bi1(m,i,j);
		}
		
		for(i=0; i<inSize; i++) {
			for(j=0; j<inSize; j++)
				GammaF0(m,i,j)=GammaF1(m,i,j);
		}
		
		for(i=0; i<inSize; i++) {
			for(j=0; j<inSize; j++)
				GammaB0(m,i,j)=GammaB1(m,i,j);
		}
		
		gamma0(m)=gamma1(m);
		
		for(i=0; i<outSize; i++)
			alfa0(m,i)=alfa1(m,i);
		
		for(i=0; i<outSize; i++) {
			for(j=0; j<inSize; j++)
				kappa0(m,i,j)=kappa1(m,i,j);
		}
	}

	// Debug to determine the lattice output
	// for(i=0; i<outSize; i++)
    //    alfa1(order,i)=d[i]-alfa1(order,i);
}

void MLSLattice::Output(val_type *out)
{
	int i;
	for(i=0; i<outSize; i++)
		// out[i]=gamma1(order)*alfa1(order,i);  this is the a posteriori error
		out[i]=alfa1(order,i);
}
