#include "global.h"
#include "system.h"
#include "assert.h"
#include <time.h>

bool MyDispatchMessage();

char aux_string[100];
int Block::number_of_blocks=0;

using namespace std;

double square(double x)
{
	return x*x;
}

double minimo(double x, double y) 
{
	if (x<y) return x;
	else return y;
}

int minimo(int x, int y) 
{
	if (x<y) return x;
	else return y;
}

double maximo(double x, double y)
{
	if (x>y) return x;
	else return y;
}

int maximo(int x, int y)
{
	if (x>y) return x;
	else return y;
}

void fft(cplx *x, int sz)
{
	int size = 1;
	while((size<<1)<=sz) size<<=1;
	ASSERT(sz==size);
	int half_size = size >> 1;

	// Bit reverse addressing
	for(int i=0, j=0; i<size; i++) {
		if (j>i) {
			cplx aux = x[i];
			x[i]=x[j]; x[j]=aux;
		};
		// add one in bit reverse addressing
		int n = half_size;
		while (j & n) { 
			j&=~n; n>>=1;
		};
		j |= n;
	}
	// ...
	for(int half_block_size = 1, num_blocks = half_size; 
		half_block_size < size; 
		half_block_size<<=1, num_blocks >>= 1)
	{
		// Twiddle factor calculation multiplier (W_N^r)
		cplx Tw_mult = cplx(cos(PI/half_block_size), -sin(PI/half_block_size));
		int i = 0; // for x[i]	
		for(int block = 0; block < num_blocks; block++, i+=half_block_size) {
			int r = 0; // for W^r_N - Actually r is not required!
			cplx Tw = cplx(1,0);
			for(int l = 0;  l < half_block_size;
				l++, i++, r+=num_blocks, Tw*=Tw_mult) {
				Tw = cplx(cos(r*2*PI/size), -sin(r*2*PI/size));
				// Butterfly
				// W_N^r = exp(-j*2*pi*r/N);
				cplx a = x[i+half_block_size]*Tw; 
				// W_N^r * x[i+B/2]
				cplx xi = x[i];
				x[i] = xi + a;
				// x[i] = x[i] + W_N^r * x[i+B/2]
				x[i+half_block_size] = xi - a;
				// x[n+B/2] = x[i] - W_N^r * x[i+B/2]
			}
		}
	}
}

void fft(val_type *in, cplx *out, int sz)
{
	for(int i=0; i<sz; i++) {
		out[i] = in[i];
	}
	fft(out, sz);
};

inline double hamming_window(int i, int order) {
	return (0.54 - 0.46*cos(2.0*PI*i/order));
}

void calc_psd(val_type *in, val_type *out, int size)
{
	cplx *aux = new cplx[size];
	int order = size-1;
	double hamming_window_sum = 0;

	for(int i=0; i<size; i++) {
		aux[i] = in[i]*hamming_window(i, order);
		hamming_window_sum+=hamming_window(i, order);
	}
	fft(aux, size);

	for(int i=0; i<size; i++) {
		out[i] = 10*log10(norm(aux[i])/hamming_window_sum+1e-10);
	}
	delete aux;
}

void fft_error(cplx *x, int sz)
{
	// DFT
	cplx *result = new cplx[sz];
	for(int k=0; k<sz; k++) {
		result[k]=0;
		for(int n=0; n<sz; n++) {
			cplx wn(cos(2*PI*k*n/sz),-sin(2*PI*k*n/sz));
			result[k]+=x[n]*wn;
		}
	}

	// FFT
	fft(x,sz);

	// Error
	for(int i=0; i<sz; i++)	x[i] -= result[i];

	delete [] result;
}

int get_n_channels(Tensor *x)
{
	int chn = 1;

	for(int j=1; j < x->dim; j++)
		chn *= x->getSize(j);
	return chn;
}

ostream & operator <<(ostream &out, Tensor *t) {
	int n_ch = get_n_channels(t);

	for(int j=0; j<t->getSize(0); j++) {
		out<<j<<" ";
		for(int i=0; i < n_ch; i++) {
			out<<*(t->values+i*t->getSize(0)+j)<<" ";
		}
		out<<endl;
	}
	return out;
}

SubSystem::SubSystem()
{
	num_blocks = 0;
	max_num_blocks=MAX_NUM_BLOCK;
	blocks= new BlockPointer[max_num_blocks];
	links_to_block = new bool[max_num_blocks];
}

SubSystem::~SubSystem()
{
      ANCDebugOutput("Destroying subsystem\n");
	  
      for(int i=0; i<num_blocks; i++) {
		  if(!links_to_block[i]) {
              delete (blocks[i]);
		  }
      }
	  delete [] blocks;
	  delete [] links_to_block;
}

void SubSystem::calc_output()
{
	for(int i=0; i<num_blocks; i++)
		blocks[i]->calc_output();
}

void SubSystem::clock_tick()
{
	for(int i=num_blocks-1; i>=0; i--)
		blocks[i]->clock_tick();
}

void SubSystem::add(BlockPointer b)
{
	ASSERT(num_blocks<max_num_blocks);
	links_to_block[num_blocks]=false;
	blocks[num_blocks++]=b;
}

void SubSystem::add2(BlockPointer b)
{
	ASSERT(num_blocks<max_num_blocks);
	links_to_block[num_blocks]=true;
	blocks[num_blocks++]=b;
}


void SubSystem::save()
{
	for(int i=0; i<num_blocks; i++)
		blocks[i]->save();
}

void SubSystem::list()
{
	for(int i=0; i<num_blocks; i++){
        sprintf(aux_string, "%d: %s\n",i,blocks[i]->getName());
		ANCDebugOutput(aux_string);
	}
}

SubSampledSystem::SubSampledSystem(int sp)
{
	sample_time = sp;
	current_time = 0;
	enable = true;
}

void SubSampledSystem::calc_output()
{
	if(current_time==0 && enable)
		SubSystem::calc_output();
}

void SubSampledSystem::clock_tick()
{
	if(current_time==0 && enable)
		SubSystem::clock_tick();

	current_time++;
	if(current_time>=sample_time)
		current_time = 0;
}

// Buffer
void Buffer::clock_tick()
{
	memmove(buf->values+1, buf->values, (buf->getSize()-1)*sizeof(val_type));
}

void Buffer::calc_output()
{
	for(int i=0; i<input->getSize(); i++)
		*(buf->values+i*buf->getSize(0))=input->values[i];
}

void MFIR::LowPass(double wc) {
	ASSERT(coeficients->dim == 1);
	int order = coeficients->getSize(0)-1;
	// Hamming window
	for(int i=0; i<=order; i++) {
		coeficients->values[i] = 0.54 - 0.46*cos(2.0*PI*i/order);
		if (i!=order/2)
			coeficients->values[i] *= sin(wc*(i-order/2))/(PI*(i-order/2));
		else coeficients->values[i] *= wc/PI;
	}

	// Calculate DC gain;
	double dc=0;
	for(int i=0; i<order+1; i++) dc+=coeficients->values[i];

	// Set  DC gain to unity
	for(int i=0; i<order+1; i++) coeficients->values[i]/=dc;
}

void SubSampledFIR(Tensor *fir, Tensor *anti_aliasing, Tensor *reconstruction, 
						int sampling_time, Tensor *result)
{
	// indices in fir and result: order(0), in(1), out(2)

	//dimensions
	ASSERT(fir->getDim()==3 && anti_aliasing->getDim()==2 
		&& reconstruction->getDim()==2 && result->getDim()==3);
		
	ASSERT(fir->getSize(1)==result->getSize(1));
	ASSERT(fir->getSize(2)==result->getSize(2));

	ASSERT(fir->getSize(1)==reconstruction->getSize(1));
	ASSERT(fir->getSize(2)==anti_aliasing->getSize(1));

	//order
	ASSERT(result->getSize(0)*sampling_time==(fir->getSize(0)+anti_aliasing->getSize(0)+
		reconstruction->getSize(0)-3+1));

	for(int l=0; l<fir->getSize(2); l++) {
		for(int r=0; l<fir->getSize(1); r++) {
	
			// subf_{k} = subf_{n} = sum_{i,j} aa_{j}*f_{i-j}*rr_{n-i}; n=freq*k;
 			for(int n=0, k=0; k<result->getSize(0); k++, n+=sampling_time) {
				result->pos(k,l,r)=0;
				for(int j=0; j<anti_aliasing->getSize(0); j++) {
					for(int i=maximo(j, n-reconstruction->getSize(0)+1); 
						i<fir->getSize(0)+j && i <= n; i++) {
						ASSERT(j>=0 && j<anti_aliasing->getSize(0));
						ASSERT(i-j>=0 && i-j<fir->getSize(0));
						ASSERT(n-i>=0 && n-i<reconstruction->getSize(0));
						result->pos(k,l,r) +=	
							sampling_time*anti_aliasing->pos(j,l,r)*
							fir->pos(i-j,l,r)*
							reconstruction->pos(n-i,l,r);
					}
				}
			}

		}
	}
}

void SubSampledFIRVector(Tensor *fir, Tensor *anti_aliasing, Tensor *reconstruction, 
						 int sampling_time, Tensor *result)
{
	// indices in fir and result: order(0), in(1), out(2)

	//dimensions
	ASSERT(fir->getDim()==1 && anti_aliasing->getDim()==1
		&& reconstruction->getDim()==1 && result->getDim()==1);

	//order
	ASSERT((result->getSize(0)-1)*sampling_time==(fir->getSize(0)+anti_aliasing->getSize(0)+
		reconstruction->getSize(0)-3));

	// subf_{k} = subf_{n} = sum_{i,j} aa_{j}*f_{i-j}*rr_{n-i}; n=freq*k;
	for(int n=0, k=0; k<result->getSize(0); k++, n+=sampling_time) {
		result->pos(k)=0;
		for(int j=0; j<anti_aliasing->getSize(0); j++) {
			for(int i=maximo(j, n-reconstruction->getSize(0)+1); 
				i<fir->getSize(0)+j && i <= n; i++) {
					ASSERT(j>=0 && j<anti_aliasing->getSize(0));
					ASSERT(i-j>=0 && i-j<fir->getSize(0));
					ASSERT(n-i>=0 && n-i<reconstruction->getSize(0));
					result->pos(k) +=	
						sampling_time*anti_aliasing->pos(j)*
						fir->pos(i-j)*
						reconstruction->pos(n-i);
				}
		}
	}
}

void FIR::SubSampledFIR(FIR *fir, MFIR *anti_aliasing, MFIR *reconstruction, 
						 int sampling_time, Tensor *in)
{
	ASSERT(coeficients == NULL);
	bool delete_fir = false;

	if(fir==NULL) {
		fir = new FIR(1, anti_aliasing->coeficients->getSize(1), 
			anti_aliasing->coeficients->getSize(1), in);
		delete_fir = true;

		for(int i=0; i<anti_aliasing->coeficients->getSize(1); i++)
			fir->coeficients->pos(0,i,i)=1;
	}

	int size = (fir->coeficients->getSize(0)+anti_aliasing->coeficients->getSize(0)+
		reconstruction->coeficients->getSize(0)-3+1)/sampling_time;

	buffer = new Buffer(size, in);
	coeficients = new Tensor(size, fir->coeficients->getSize(1),
		fir->coeficients->getSize(2));
	output = new Tensor(fir->coeficients->getSize(1));
	link = false;

	::SubSampledFIR(fir->coeficients, anti_aliasing->coeficients, 
		reconstruction->coeficients, sampling_time, coeficients);

	if(delete_fir) delete fir;
}

void MFIR::SubSampledFIR(MFIR *fir, MFIR *anti_aliasing, MFIR *reconstruction, 
						 int sampling_time, Tensor *in)
{
	ASSERT(coeficients == NULL);
	bool delete_fir = false;

	if(fir==NULL) {
		fir = new MFIR(1, in);
		fir->coeficients->pos(0)=1;
	}

	int size = (fir->coeficients->getSize(0)+anti_aliasing->coeficients->getSize(0)+
		reconstruction->coeficients->getSize(0)-3)/sampling_time+1;

	buffer = new Buffer(size, in);
	coeficients = new Tensor(size);
	output = new Tensor(in->getSize(0));
	link = false;

	if (delete_fir) delete fir;

	::SubSampledFIRVector(fir->coeficients, anti_aliasing->coeficients, 
		reconstruction->coeficients, sampling_time, coeficients);
}

int dsp_sampling_time;

ANC::ANC()
{
	name = "ANC main System";
	double analog_sampling_time = 1.0/12000; //12kHz
	double usefull_band = 1000; // Hz (341 in PSD plot)
	dsp_sampling_time = 3; // 4kHz (in analog_sampling_time units!)

	number_reference_sensors=2;  //2
	number_anti_noise_sources=3; //3
	number_error_sensors=2;      //2

	int currente_time = 0;
	int filter_size = 129;
	int anti_aliasing_filter_size = 60;
	int anti_noise_filter_size = 40;
	int inversion_delay_value = 50;

	add(anc_dsp = new SubSampledSystem(dsp_sampling_time));
	anc_dsp->name="anc_dsp";
	add(off_line_modelling_dsp = new SubSampledSystem(dsp_sampling_time));
	off_line_modelling_dsp->name ="off_line_modelling_dsp";

	add(reference = new WhiteNoiseGenerator(1, number_reference_sensors));
	reference->name="reference";

	add(primary_path = new FIR(300, number_reference_sensors,
		number_error_sensors, reference->output, "Primary_Path"));

	for(int i=0; i<number_reference_sensors; i++) {
		for(int j=0; j<number_error_sensors; j++) {
			int ind[3];
			ind[0] = (int)(200+50.0*rand()/RAND_MAX); ind[1] = i; ind[2] = j;
			primary_path->coeficients->pos(ind) = 1;
		}
	}

	add(reference_anti_aliasing_filter_bank = new MFIR(anti_aliasing_filter_size , reference->output));
	reference_anti_aliasing_filter_bank->LowPass(PI/dsp_sampling_time);

	anc_dsp->add(controller = new LatticeBlk(reference_anti_aliasing_filter_bank->output, number_anti_noise_sources));
	controller->name = "controller";

	anc_dsp->add(gain1 = new Gain(controller->output, -1) );

	anc_dsp->add(aditive_noise = new WhiteNoiseGenerator(1,number_anti_noise_sources));
	off_line_modelling_dsp->add2(aditive_noise);
	aditive_noise->name="aditive_noise";

	anc_dsp->add(additive_noise_sum = new Sum( gain1->output, aditive_noise->output));
	additive_noise_sum->name = "additive_noise_sum";
	off_line_modelling_dsp->add2(additive_noise_sum);

	add(anti_noise_reconstruction_filter_bank = new RFIR(anti_aliasing_filter_size ,
		additive_noise_sum->output, dsp_sampling_time));
	anti_noise_reconstruction_filter_bank->LowPass(PI/dsp_sampling_time);
	anti_noise_reconstruction_filter_bank->name="anti_noise_reconstruction_filter_bank";

	add(cancelation_path = new FIR(filter_size, number_anti_noise_sources,
		number_error_sensors, anti_noise_reconstruction_filter_bank->output));
	cancelation_path->name="cancelation_path";

	for(int i=0; i<number_anti_noise_sources; i++) {
		for(int j=0; j<number_error_sensors; j++) {
			int ind[3];
			ind[0] = (int)(10+20.0*rand()/RAND_MAX); ind[1] = i; ind[2] = j;
			cancelation_path->coeficients->pos(ind) = 1;
		}
	}
	add(acoustic_sum = new Sum(primary_path->output, cancelation_path->output));

	add(error_anti_aliasing_filter_bank = new MFIR(anti_aliasing_filter_size, acoustic_sum->output));
	error_anti_aliasing_filter_bank->LowPass(PI/dsp_sampling_time);
	
	anc_dsp->add(inversion_delay = new Delay(inversion_delay_value, additive_noise_sum->output));
	off_line_modelling_dsp->add2(inversion_delay);
	inversion_delay->name = "inversion_delay";

	anc_dsp->add(inverse_cancelation_path_estimate = 
		new LatticeBlk(error_anti_aliasing_filter_bank->output, number_anti_noise_sources));
	inverse_cancelation_path_estimate->name="inverse cancelation path estimate";
	off_line_modelling_dsp->add2(inverse_cancelation_path_estimate);

	off_line_modelling_dsp->add(inverse_cancelation_path_estimate0 = 
		new LSLBlk(filter_size, number_error_sensors, number_anti_noise_sources,
			error_anti_aliasing_filter_bank->output, inversion_delay->output));
	inverse_cancelation_path_estimate->Link(inverse_cancelation_path_estimate0);
	inverse_cancelation_path_estimate0->name="inverse_cancelation_path_estimate0";
	
	anc_dsp->add(dif1 = new Difference(inverse_cancelation_path_estimate->output, inversion_delay->output));
	off_line_modelling_dsp->add2(dif1);

	anc_dsp->add(inversion_delay2 = new Delay(inversion_delay_value, 
		reference_anti_aliasing_filter_bank->output));

	anc_dsp->add(anti_noise_filter_delay = new Delay((anti_noise_filter_size-1)/2, 
		inversion_delay2->output));

	anc_dsp->add(anti_noise_filter = new MFIR(anti_noise_filter_size, dif1->output));
	anti_noise_filter->LowPass(2*PI*usefull_band*analog_sampling_time*dsp_sampling_time);
	anti_noise_filter->name="anti_noise_filter";

	anc_dsp->add(lsl1 = new LSLBlk(filter_size, number_reference_sensors, number_anti_noise_sources,
		anti_noise_filter_delay->output, anti_noise_filter->output));
	lsl1->name="lsl1";

	controller->Link(lsl1);
}

void ANC::Simulate()
{		
	float real_time;
    	srand( (unsigned)time( NULL ));

	int start_time = clock();

	off_line_modelling_dsp->enable = true;
	anc_dsp->enable = false;
	reference->amplitude = 0;
	aditive_noise->amplitude = 1;
      
      //return;

	for(currente_time=0; currente_time<30000; currente_time++) {
		real_time = (float)(1.0*clock()-start_time)/CLOCKS_PER_SEC;

		if (currente_time == 10000) {
			reference->amplitude = 1;
			aditive_noise->amplitude = 0.01;
			off_line_modelling_dsp->enable = false;
			anc_dsp->enable = true;
		}

		calc_output();
		clock_tick();

		if ( !(currente_time % 1000) ) {
                        printf("time: %d\n", currente_time);
			/*
			for(int i=0; i<number_anti_noise_sources; i++){
				printf("residual %d is : %f\n", i, 
					lsl1->output->pos(i));
			}*/
		}

	}
	
	printf("Frequency: %f (Hz)\n",currente_time/real_time/dsp_sampling_time); 

	save();
}
